//  Data Structures with Java by John R. Hubbard
//  Copyright McGraw-Hill, 2001
//  Example 12.9 on page 232
//  An IntPriorityQueue class

public class IntPriorityQueue implements IntPriorityQueueInterface
{ private final int CAPACITY=16;
  private int capacity, size;
  private int[] ints;
  
  public IntPriorityQueue()
  { capacity = CAPACITY;
    ints = new int[capacity+1];
  }
  
  public IntPriorityQueue(int capacity)
  { this.capacity = capacity;
    ints = new int[capacity+1];
  }
  
  public int top()
  { return ints[1];
  }
  
  public int pop()
  { int x = ints[1];
    ints[1] = ints[size--];
    heapifyDown();
    return x;
  }
  
  public void push(int x)
  { if (size==capacity) resize();
    ints[++size] = x;
    heapifyUp();
  }
  
  public int size()
  { return size;
  }
  
  public String toString()
  { if (size==0) return "[]";
    String string = "[" + ints[1];
    for (int i=2; i<=size; i++)
      string += "," + ints[i];
    return string + "]";
  }
  
  private void heapifyUp()
  { for (int j=size, i; j>0; j = i)
    { i = j/2;
      if (ints[j]<ints[i]) swap(ints,i,j);
    }
  }
    private void heapifyDown()
  { for (int i=1, j; i<=size/2; i = j)
    { j = 2*i;
      if (j<size && ints[j]>ints[j+1]) ++j;
      if (ints[j]<ints[i]) swap(ints,i,j);
    }
  }
  
  private void resize()
  { int[] temp = new int[capacity+1];
    for (int i=1; i<=capacity; i++)
      temp[i] = ints[i];
    capacity *= 2;
    ints = new int[capacity+1];
    for (int i=1; i<=size; i++)
      ints[i] = temp[i];
  }
  
  private void swap(int[] a, int i, int j)
  { int tmp=a[i];
    a[i] = a[j];
    a[j] = tmp;
  }
} 
