//  Data Structures with Java by John R. Hubbard
//  Copyright McGraw-Hill, 2001
//  Problem 12.8 on page 239
//  Using the PriorityQueue class in the client/server simulation

import schaums.dswj.*;
import schaums.dswj.PriorityQueue;

public class Pr1208
{ private static final int NUMBER_OF_SERVERS = 4;
  private static final double MEAN_INTERARRIVAL_TIME = 5.0;
  private static final int DURATION = 200;
  private static Server[] servers = new Server[NUMBER_OF_SERVERS];
  private static PriorityQueue clients
    = new PriorityQueue(new ClientComparator());
  private static Random random
   = new Random(MEAN_INTERARRIVAL_TIME);

  public static void main(String[] args)
  { for (int i=0; i<NUMBER_OF_SERVERS; i++)
      servers[i] = new Server();
    int timeOfNextArrival = random.intNextExponential();
    for (int t=0; t<DURATION; t++)
    { if (t == timeOfNextArrival)  // new arrival
      { clients.push(new Client(t));
        timeOfNextArrival += random.intNextExponential();
        System.out.println(t + "\t" + clients);
      }
      for (int i=0; i<NUMBER_OF_SERVERS; i++)  // process servers
        if (servers[i].isFree())
        { if(!clients.isEmpty())
            servers[i].beginServing((Client)clients.pop(),t);
        }
        else if (t == servers[i].getTimeServiceEnds())
          servers[i].endServing(t);
    }
  }
}

public class Server
{ private static Random randomMeanServiceRate
    = new Random(1.00,0.20);
  private static char nextId = 'A';
  private Random randomServiceRate;
  private char id;
  private double meanServiceRate, serviceRate;
  private Client client;
  private int timeServiceEnds;

  public Server()
  { id = (char)nextId++;
    meanServiceRate = randomMeanServiceRate.nextGaussian();
    randomServiceRate = new Random(meanServiceRate,0.10);
  }

  public Client getClient()
  { return client;
  }

  public void beginServing(Client client, int time)
  { this.client = client;
    serviceRate = randomServiceRate.nextGaussian();
    client.beginService(this,time);
    int serviceTime
      = (int)Math.ceil(client.getJobSize()/serviceRate);
    timeServiceEnds = time + serviceTime;
    System.out.println(time + "\t" + id + " starts serving "
                            + client);
  }

  public void endServing(int time)
  { client.endService(time);
    System.out.println(time + "\t" + id + " finishes serving "
                            + client);
    this.client = null;
  }

  public int getTimeServiceEnds()
  { return timeServiceEnds;
  }

  public boolean isFree()
  { return client == null;
  }

  public String toString()
  { int percentMeanServiceRate
          = (int)Math.round(100*meanServiceRate);
    int percentServiceRate = (int)Math.round(100*serviceRate);
    return id + "(" + percentMeanServiceRate + "%,"
              + percentServiceRate + "%)";
  }
}

public class ClientComparator implements java.util.Comparator
{ public int compare(Object o1, Object o2)
  { if (o1==null || !(o1 instanceof schaums.dswj.Client))
      return -1;
    Client c1 = (Client)o1;
    if (o2==null || !(o2 instanceof Client)) return  1;
    Client c2 = (Client)o2;
    if (c1.getJobSize() < c2.getJobSize()) return -1;
    if (c1.getJobSize() > c2.getJobSize()) return  1;
    return 0;
  }
}

public class Client
{ public static final int MEAN_JOB_SIZE = 20;
  private static Random randomJobSize = new Random(MEAN_JOB_SIZE);
  private static int nextId = 0;
  private int id, jobSize, tArrived, tBegan, tEnded;
  private Server server;

  public Client(int time)
  { id = ++nextId;
    jobSize = randomJobSize.intNextExponential();
    tArrived = time;
    System.out.println(time + "\t" + this + " arrives");
  }

  public double getJobSize()
  { return jobSize;
  }
  public int getWaitTime()
  { return tBegan - tArrived;
  }

  public int getServiceTime()
  { return tEnded - tBegan;
  }

  public void beginService(Server server, int time)
  { this.server = server;
    tBegan = time;
  }

  public void endService(int time)
  { tEnded = time;
    server = null;
  }

  public String toString()
  { return "#" + id + "(" + (int)Math.round(jobSize) + ")";
  }

  private static void print(int job, int time, double size)
  { System.out.println("Job #" + job + " arrives at time " + time
      + " with " + (int)Math.round(size) + " pages.");
  }

  private static void printBegins(Server server, int job, int time)
  { System.out.println("Printer " + server + " begins Job #" + job
      + " at time " + time + ".");
  }

  private static void printEnds(Server server, int job, int time)
  { System.out.println("Printer " + server + " ends Job #" + job
      + " at time " + time + ".");
  }
}
